/*
 * Copyright (C) 2012 devesion.com
 * 
 * Contact: Lukas Dembinski <dembol@devesion.com>
 * Initial developer: Lukas Dembinski <dembol@devesion.com>
 * Contributor(s):
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

package com.devesion.n2io.connection.channel;

import com.devesion.n2io.connection.ConnectionException;
import com.devesion.n2io.connection.selectable.AbstractSelectableConnection;
import java.io.IOException;
import java.nio.channels.SelectableChannel;
import java.util.Set;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author dembol
 */
public abstract class AbstractSelectableChannelConnection extends AbstractSelectableConnection {

    private static final Logger LOGGER = LoggerFactory.getLogger(AbstractSelectableChannelConnection.class);
    
    private Dispatcher dispatcher;
    
    private final SelectableChannel channel;
    
    private boolean closedFlag = false;
    
    
    public AbstractSelectableChannelConnection(SelectableChannel channel) {
        this.channel = channel;
    }

    public SelectableChannel getChannel() {
        return channel;
    }

    @Override
    public void close() throws ConnectionException {
        
        if (isClosed()) {
            return;
        }
        
        try {
            cancelGenerateEvents();
            getChannel().close();
            setClosedFlag(true);
        } catch (IOException iOException) {
            throw new ConnectionException(iOException);
        }
    }

    void setClosedFlag(boolean closedFlag) {
        this.closedFlag = closedFlag;
    }
    
    @Override
    public boolean isClosed() {
        return !getChannel().isOpen() | closedFlag;
    }
    
    @Override
    public void configureBlocking(boolean blocking) throws ConnectionException {

        if (isClosed()) {
            return;
        }

        try {
            getChannel().configureBlocking(blocking);
        } catch (IOException iOException) {
            throw new ConnectionException("cannot set " + (blocking ? "" : "non ") + "blocking mode for connection", iOException);
        }
    }

    private Dispatcher getDispatcher() throws ConnectionException {
        
        if (dispatcher == null) {
            dispatcher = DispatcherBalancer.getInstance().getNextDispatcher();
        }
        
        return dispatcher;
    }
    
    @Override
    public void startGenerateEvents(Set<String> allowedEvents) throws ConnectionException {
        getDispatcher().addConnection(this, allowedEvents);
    }

    @Override
    public void stopGenerateEvents() throws ConnectionException {
        if (dispatcher != null) {
            dispatcher.removeConnection(this);
        }
    }

    @Override
    public void cancelGenerateEvents() throws ConnectionException {
        getDispatcher().cancelConnection(this);
    }
}
