/*
 * Copyright (C) 2012 devesion.com
 * 
 * Contact: Lukas Dembinski <dembol@devesion.com>
 * Initial developer: Lukas Dembinski <dembol@devesion.com>
 * Contributor(s):
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

package com.devesion.n2io.connection.channel;

import com.devesion.n2io.connection.ConnectionException;
import com.devesion.n2io.connection.selectable.Event;
import java.io.IOException;
import java.nio.channels.CancelledKeyException;
import java.nio.channels.SelectableChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.util.Set;
import lombok.EqualsAndHashCode;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Event generator.
 *
 * @author dembol
 */
@EqualsAndHashCode
class Dispatcher implements Runnable, Comparable<Dispatcher> {

    private static final Logger LOGGER = LoggerFactory.getLogger(Dispatcher.class);
   
    /**
     * Keys selector.
     */
    private final Selector sel;
    
    /**
     * Guard object.
     */
    private final Object guard = new Object();

    /**
     * Constructor.
     *
     * @param queue event queue
     * @throws IOException
     */
    public Dispatcher() throws ConnectionException {
                
        try {
            sel = Selector.open();
        } catch (IOException ex) {
            throw new ConnectionException("cannot open async selector", ex);
        }
        
        init();
    }

    public final void addConnection(AbstractSelectableChannelConnection conn, Set<String> allowedEvents) throws ConnectionException {
        
        if (conn.isClosed()) {
            return;
        }

        final SelectableChannel channel = conn.getChannel();
        try {
            synchronized (guard) {
                sel.wakeup();
                
                int interestsOps = 0;
                for (String eventName : allowedEvents) {
                    if (Event.EVENT_READ.equals(eventName)) {
                        interestsOps |= SelectionKey.OP_READ;
                    } else
                    if (Event.EVENT_WRITE.equals(eventName)) {
                        interestsOps |= SelectionKey.OP_WRITE;
                    } else {
                        LOGGER.error("unknown event name '" + eventName + "'");
                        throw new ConnectionException("Unknown event name '" + eventName + "'");
                    }
                }
                
                if (!channel.isRegistered()) {
                    channel.register(sel, interestsOps, conn);
                } else {
                    try {
                        final SelectionKey key = channel.keyFor(sel);
                        if (key != null) {
                            key.interestOps(interestsOps);
                        } else {
                            sel.selectNow();
                            channel.register(sel, interestsOps, conn);
                        }
                    } catch (CancelledKeyException e) {
                        sel.selectNow();
                        channel.register(sel, interestsOps, conn);
                    }
                }
            }

        } catch (Exception e) {
            throw new ConnectionException("cannot add asynchronous context to dispatcher", e);
        }
    }

    public final void removeConnection(AbstractSelectableChannelConnection conn) {
        
        if (conn.isClosed()) {
            return;
        }

        synchronized (guard) {
            sel.wakeup();
            final SelectableChannel channel = conn.getChannel();
            final SelectionKey key = channel.keyFor(sel);
            if ((key != null) && (key.isValid())) {
                key.interestOps(0);
            }
        }
    }

    public final void cancelConnection(AbstractSelectableChannelConnection conn) {
        
        if (conn.isClosed()) {
            return;
        }

        synchronized (guard) {
            sel.wakeup();
            final SelectableChannel channel = conn.getChannel();
            final SelectionKey key = channel.keyFor(sel);
            if (key != null) {
                key.cancel();
            }
        }
    }
    
    /**
     * {@inheritDoc}
     *
     * @see java.lang.Runnable#run()
     */
    @SuppressWarnings("empty-statement")
    @Override
    public void run() {

        while (true) {

            try {
                synchronized (guard) { };
                
                if (sel.select() <= 0) {
                    continue;
                }
            } catch (IOException e) {
                continue;
            }

            synchronized(guard) {
            for (SelectionKey k : sel.selectedKeys()) {
                if (k.isValid() && (k.attachment() != null)) {
                    final AbstractSelectableChannelConnection conn = (AbstractSelectableChannelConnection) k.attachment();
                    
                    Event e;
                    if (k.isReadable()) {
                        e = new Event(Event.EVENT_READ, conn);
                    } else if (k.isWritable()) {
                        e = new Event(Event.EVENT_WRITE, conn);
                    } else {
                        continue;
                    }

                    removeConnection(conn);
                    conn.handleEvent(e);
                }
            }
            }

            sel.selectedKeys().clear();
        }
    }

    private void init() {
        final Thread t = new Thread(this);
        t.start();
    }

    @Override
    public int compareTo(Dispatcher o) {
        if (o.equals(this)) {
            return 0;
        }
        
        return 1;
    }
}
