/*
 * Copyright (C) 2012 devesion.com
 * 
 * Contact: Lukas Dembinski <dembol@devesion.com>
 * Initial developer: Lukas Dembinski <dembol@devesion.com>
 * Contributor(s):
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

package com.devesion.n2io.server;

import com.devesion.n2io.N2IOException;
import com.devesion.n2io.connection.Connection;
import com.devesion.n2io.server.monitor.N2ioServerMonitor;
import com.devesion.n2io.stage.AbstractStage;
import com.devesion.n2io.stage.StageContext;
import com.devesion.n2io.stage.StageException;
import com.devesion.n2io.stage.StageRouting;
import java.lang.management.ManagementFactory;
import java.util.HashMap;
import java.util.Map;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * AbstractServer.
 * 
 * Main responsibilities:
 * - stage routing table managing
 * - accepting connections
 * - enqueuing connections in the first stage
 * 
 * @author dembol
 */
public abstract class AbstractServer {

    private static final Logger LOGGER = LoggerFactory.getLogger(AbstractServer.class);
    
    private final Map<String, AbstractStage> stages = new HashMap<String, AbstractStage>();
    
    private final StageRouting routing = new StageRouting();
 
    public final void registerStage(AbstractStage s) {
        stages.put(s.getName(), s);
    }
    
    protected AbstractStage findStage(String name) throws StageException {
        
        final AbstractStage stage = stages.get(name);
        if (stage == null) {
            throw new StageException("stage '" + name + "' not found");
        }
        
        return stage;
    }
    
    /**
     * Add route from some stage to the other stage.
     * 
     * @param from source stage
     * @param to destination stage
     * @param ruleName name of the rule (unused in this moment)
     * @throws StageException 
     */
    public final void addRoute(String from, String to, String ruleName) throws StageException {
        
        final AbstractStage fromStage = findStage(from);
        final AbstractStage toStage = findStage(to);
        routing.addRoute(fromStage, toStage, ruleName);
    }

    /**
     * Add route from some stage to the other stage.
     * 
     * @param from source stage
     * @param to destination stage
     * @throws StageException 
     */
    public final void addRoute(String from, String to) throws StageException {
        
        final AbstractStage fromStage = findStage(from);
        final AbstractStage toStage = findStage(to);
        routing.addRoute(fromStage, toStage);
    }
    
    private void registerMBean() throws ServerException {
        
        final MBeanServer mbs = ManagementFactory.getPlatformMBeanServer();
        
        try {
            mbs.registerMBean(new N2ioServerMonitor(), new ObjectName("pl.sysdesign.n2io:type=Server"));
        } catch (Exception e) {
            throw new ServerException("cannot register server MBean", e);
        }
    }
    
    protected abstract Connection accept() throws ServerException;
    
    protected abstract void init() throws ServerException;

    /**
     * The main method of the server.
     * 
     * @param startStageName name of the first stage
     * @throws N2IOException 
     */
    public void start(String startStageName) throws N2IOException {

        LOGGER.info("Starting server");
        
        registerMBean();
 
        init();
        
        for (AbstractStage s : stages.values()) {
            s.init();
        }
        
        for (;;) {
            final Connection c = accept();
            final StageContext ctx = new StageContext(routing, c);
            final AbstractStage startStage = findStage(startStageName);
            
            if (startStage == null) {
                throw new ServerException("cannot find start stage '" + startStageName + "'");
            }
            
            startStage.enqueue(ctx);
        }
    }
}
