/*
 * Copyright (C) 2012 devesion.com
 * 
 * Contact: Lukas Dembinski <dembol@devesion.com>
 * Initial developer: Lukas Dembinski <dembol@devesion.com>
 * Contributor(s):
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

package com.devesion.n2io.stage;

import com.devesion.n2io.connection.Connection;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.PriorityBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * Class representing single processing stage.
 * 
 * @author dembol
 */
public abstract class AbstractStage {

    private static final int DEFAULT_THREAD_NUMBER = 1;
    
    /**
     * AbstractStage name.
     */
    protected final String name;
    
    private ExecutorService s;
    
    private final int threads;

    private final StageHandler handler;
  
    /**
     * Constructor.
     * 
     * @param name the name of the stage
     */
    public AbstractStage(String name, StageHandler handler) {
        this(name, DEFAULT_THREAD_NUMBER, handler);
    }
    
    /**
     * Constructor.
     * 
     * @param name the name of the stage
     */
    public AbstractStage(String name, int threads, StageHandler handler) {
        this.name = name;
        this.threads = threads;
                
        if (handler == null) {
            throw new RuntimeException("stage handler is null for stage '" + name + "'");
        }
        this.handler = handler;
    }

    public String getName() {
        return name;
    }

    protected StageHandler getHandler() {
        return handler;
    }
    
    public void init() {
        s = new ThreadPoolExecutor(threads, threads, 60, TimeUnit.SECONDS, new PriorityBlockingQueue<Runnable>());
    }
    
    void enqueue(StageTask task) {
        s.execute(task);
    }
    
    /**
     * Enqueue stage context.
     * 
     * @param ctx processing context
     */
    public void enqueue(StageContext ctx) {
        ctx.setStage(this);
        ctx.setHandled(false);
        enqueue(new StageTask(this, ctx));
    }

    protected boolean checkConnection(StageContext ctx) throws StageException {
  
        final Connection c = ctx.getConnection();
        if (c.isClosed()) {
            handler.onClose(ctx);
            return false;
        }
        
        return true;
    }
    
    protected abstract void process(StageContext ctx) throws StageException;
    
    protected void processError(StageContext ctx, Throwable t) throws StageException {
        getHandler().onError(ctx, t);
    }
}
