/*
 * Copyright (C) 2012 devesion.com
 * 
 * Contact: Lukas Dembinski <dembol@devesion.com>
 * Initial developer: Lukas Dembinski <dembol@devesion.com>
 * Contributor(s):
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

package com.devesion.n2io.stage;

import com.devesion.n2io.connection.Connection;
import com.devesion.n2io.connection.ConnectionException;
import lombok.EqualsAndHashCode;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * The class binds stage with some processing context.
 * 
 * @author dembol
 */
@EqualsAndHashCode(of= {"taskId", "priority" })
class StageTask implements Runnable, Comparable<StageTask> {

    private final long taskId;
    
    private final int priority;
        
    private final AbstractStage s;
   
    private final StageContext ctx;

    private static final int DEFAULT_PRIORITY = 10;
    
    private static final Logger LOGGER = LoggerFactory.getLogger(StageTask.class);
    
    static class StageTaskIdGenerator {
        
        private static long curTaskId = 0;
        
        private static synchronized long getTaskId() {
            curTaskId += 1;
            return curTaskId;
        }
    }
    
    public StageTask(AbstractStage s, StageContext ctx, int priority) {
        this.s = s;
        this.ctx = ctx;
        this.priority = priority;
        this.taskId = StageTaskIdGenerator.getTaskId();
    }

    public StageTask(AbstractStage s, StageContext ctx) {
        this(s, ctx, DEFAULT_PRIORITY);
    }
    
    public StageContext getContext() {
        return ctx;
    }
    
    protected AbstractStage getStage() {
        return s;
    }
    
    protected void processInternally() throws StageException {
        getStage().process(ctx);
    }

    protected void process() throws StageException {
        try {
            processInternally();
        } catch (RuntimeException t) {
            getStage().processError(ctx, t);
            throw t;
        } catch(StageException ex) {
            getStage().processError(ctx, ex);
            throw ex;
        }
    }
    
    @Override
    public void run() {
        
        try {
            process();
        } catch (Throwable t) {
            LOGGER.error("stage task processing error. Closing connection.", t);
            
            final Connection c = ctx.getConnection();
            if (c != null) {
                try {
                    c.close();
                } catch (ConnectionException e) {
                    LOGGER.error("cannot close connection", e);
                }
            }
        }
    }

    @Override
    public int compareTo(StageTask task) {
            
        if (this.equals(task)) {
            return this.priority - task.priority;
        } else 
        if (task.taskId < this.taskId) {
            return 1;
        }
        
        return -1;
    }
}
