/*
 * Copyright (C) 2012 devesion.com
 * 
 * Contact: Lukas Dembinski <dembol@devesion.com>
 * Initial developer: Lukas Dembinski <dembol@devesion.com>
 * Contributor(s):
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

package com.devesion.n2io.samples.http;

import com.devesion.n2io.connection.ConnectionException;
import com.devesion.n2io.connection.decorators.TextConnection;
import com.devesion.n2io.connection.selectable.Event;
import com.devesion.n2io.stage.AsyncStageHandler;
import com.devesion.n2io.stage.StageContext;
import com.devesion.n2io.stage.StageException;
import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.HashSet;
import java.util.Set;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author dembol
 */
public class SendResponseHandler extends StandardHandler implements AsyncStageHandler {
    
    private static final Logger LOGGER = LoggerFactory.getLogger(SendResponseHandler.class);
    
    private static final int DEFAULT_BUFFER_SIZE = 0x2FFF;
    
    private static Set<String> allowedEventsSet = new HashSet<String>();

    @Override
    public Set<String> allowedEvents() {
        return allowedEventsSet;
    }

    @Override
    public void onStart(StageContext ctx) throws StageException {
        allowedEventsSet.add(Event.EVENT_WRITE);
    }

    @Override
    public void onEnd(StageContext ctx) throws StageException {
        super.closeConnection(ctx);
        cleanResources(ctx);
    }

    private void cleanResources(StageContext ctx) {
        
        final BufferedInputStream bis = (BufferedInputStream) ctx.getParameter("bis");
        if (bis != null) {
            try {
                bis.close();
                ctx.setParameter("bis", null);
            } catch (IOException ex) {
                LOGGER.equals("cannot close input stream");
            }
        }
    }

    private byte[] getReadBuffer(StageContext ctx) {
                
        byte[] b = (byte[])ctx.getParameter("read-buffer");
        if (b == null) {
            b = new byte[DEFAULT_BUFFER_SIZE];
            ctx.setParameter("read-buffer", b);
        }
        
        return b;
    }
    
    private boolean sendData(BufferedInputStream bis, StageContext ctx) {
        
        final TextConnection c = new TextConnection(ctx.getConnection());
        Integer pos = (Integer) ctx.getParameter("position");
        if (pos == null) {
            pos = 0;
        }
        
        if (pos == 0) {
            try {
                c.writeText("HTTP/1.1 200 OK");
                c.writeText(Constants.CRLF);
                c.writeText(Constants.CRLF);
        
            } catch (Exception ex) {
                if (bis != null) {
                    try {
                        bis.close();
                    } catch (IOException ex2) {
                        LOGGER.error("cannot close file input stream");
                    }
                }

                LOGGER.info("cannot write headers - " + ex.getMessage());
                return true;
            }
        }

        try {
            final byte[] b = getReadBuffer(ctx);
            bis.skip(pos);
            final int len = bis.read(b);
            
            if (len >= 0) {
                final int writeLen = c.write(ByteBuffer.wrap(b, 0, len));
                pos += writeLen;
                ctx.setParameter("position", pos);
                return false;
            } 
        } catch (IOException iOException) {
            LOGGER.error("cannot read file - " + iOException.getMessage());
        } catch (ConnectionException connectionException) {
            LOGGER.error("cannot write data - " + connectionException.getMessage());
        }
        
        return true;
    }

    private boolean onWrite(StageContext ctx) throws StageException {
        
        BufferedInputStream bis = null;
        final String path = (String) ctx.getParameter("path");
        if (path != null) {
            try {
                bis = new BufferedInputStream(new FileInputStream(path));
            } catch (FileNotFoundException ex) {
                LOGGER.info("file not found");
            }
        }

        final TextConnection c = new TextConnection(ctx.getConnection());
        if (bis == null) {
            try {
                c.writeText("HTTP/1.1 404 Not Found");
                c.writeText(Constants.CRLF);
                c.writeText(Constants.CRLF);
            } catch (ConnectionException ex) {
                LOGGER.info("cannot write headers - " + ex.getMessage());
            }
           
            return true;
        } 
        
        final boolean res = sendData(bis, ctx);
        
        if (bis != null) {
            try {
                bis.close();
            } catch (IOException ex) {
                LOGGER.error("cannot close file input stream");
            }
        }
        
        
        return res;        
    }
    
    @Override
    public boolean onEvent(StageContext ctx, Event e) throws StageException {

        if (e.getName().equals(Event.EVENT_WRITE)) {
            return onWrite(ctx);
        }
        
        // something is wrong
        return true;
    }

    @Override
    public void onError(StageContext ctx, Throwable t) throws StageException {
        cleanResources(ctx);
        super.closeConnection(ctx);
    }

    @Override
    public void onClose(StageContext ctx) throws StageException {
        cleanResources(ctx);
        super.closeConnection(ctx);
    }
}
